"""

"""
from ebooklet.main import open_ebooklet, open_rcg, EVariableLengthValue, RemoteConnGroup
from ebooklet.remote import S3Connection

__all__ = ["open_ebooklet", "open_rcg", "EVariableLengthValue", 'RemoteConnGroup', 'S3Connection']

__version__ = '0.8.0'
